/**
 * Zilo - the tiny web server
 * Author:  Stefan Hellkvist, stefan@verkstad.net
 * Version: 0.92
 **/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <signal.h>
#include <ctype.h>

#define DEBUG_LEVEL 0
#define MAX_PATH 255
#define MAX_REQUEST_LENGTH 2048
#define PORT 8080
#define BUFFER_SIZE 1024
#define VERSION "0.95"
#define ERROR_PAGE "You are being silly!\n" \
                   "You were just insulted by Zilo (version " VERSION ").\n"
#if PORT != 80
#warning This webserver is not using the default http port.
#endif

inline void handleRequest( int sockfd );
inline void sendBackFile( char *requested, int client_sockfd );
inline void init( int argc, char **argv );
inline void usage( void );
inline void mainLoop( void );

static char root[MAX_PATH];
int server_sockfd, client_sockfd, client_len;
struct sockaddr_in server_address, client_address;

int
main( int argc, char **argv )
{
    init( argc, argv );
    mainLoop();
    return 0;
}

inline void
mainLoop()
{
    while ( 1 )
    {
	client_sockfd = accept( server_sockfd, 
				(struct sockaddr *) &client_address,
				&client_len );
#ifdef MULTI_THREADED
	if ( !fork() )
	{
#endif
	    handleRequest( client_sockfd );
#ifdef MULTI_THREADED
	    exit( 0 );
	}
	else
	    close( client_sockfd );
#endif
    }
}

inline void 
init( int argc, char **argv )
{
    if ( argc != 2 )
    {
	usage();
        exit( 0 );
    }

    if ( strlen( argv[1] ) > MAX_PATH - 1 )
    {
	fprintf( stderr, "Path to root directory is too long\n" );
	exit( 2 );
    }
    strcpy( root, argv[1] );

#ifdef MULTI_THREADED
    signal( SIGCHLD, SIG_IGN );
#endif
    server_sockfd = socket( AF_INET, SOCK_STREAM, 0 );
    if ( server_sockfd == -1 )
    {
	perror( "zilo: " );
	exit( 3 );
    }
    server_address.sin_family = AF_INET;
    server_address.sin_addr.s_addr = htonl(INADDR_ANY);
    server_address.sin_port = htons(PORT);

    if ( bind( server_sockfd, (struct sockaddr *) &server_address, 
	       sizeof(server_address) ) == -1 )
    {
	perror( "zilo: Could not bind." );
	exit( 4 );
    }

    if ( listen( server_sockfd, 5 ) == -1 )
    {
	perror( "zilo: " );
	exit( 5 );
    }

    printf( "Zilo version %s started\n", VERSION );
#if PORT != 80
    printf( "Listening to port: %d\n", PORT );
#endif
    daemon( 1, 1 );
}

inline void
usage( void )
{
    printf( "usage: zilo <root-directory>\n" );
}

inline void
handleRequest( int client_sockfd )
{
    char buffer[MAX_REQUEST_LENGTH], filename[MAX_PATH];
    int len;
    
    if ( ( len = recv( client_sockfd, buffer, MAX_REQUEST_LENGTH, 0 )) == -1 )
	return;
    buffer[ len - 1 ] = '\0';
    buffer[ 3 ] = '\0';
    if ( strcmp( "GET", buffer ) != 0 )
    	write( client_sockfd, ERROR_PAGE, strlen( ERROR_PAGE ) );
    else
    {
	char *requested = buffer + 4;
	char *ptr = requested;

	// Remove everything after the filename
	while ( *ptr )
	{
	    if ( isspace( *ptr ) )
	    {
                *ptr = '\0';
                break;
            }
	    ptr++;
	}
	strcpy( filename, root );
	strcat( filename, requested );
#if DEBUG_LEVEL > 0
	printf( "GET %s\n", requested );
#endif
	if ( filename[ strlen( filename ) - 1 ] == '/' ) 
	    strcat( filename, "index.html" );
	sendBackFile( filename, client_sockfd );
    }
    close( client_sockfd );
}

inline void 
sendBackFile( char *filename, int client_sockfd )
{
    FILE *fp;
    char buffer[BUFFER_SIZE];
    size_t len;

    fp = fopen( filename, "r" );
    if ( fp == NULL || strstr( filename, ".." ) )
    {
	write( client_sockfd, ERROR_PAGE, strlen( ERROR_PAGE ) );
	if ( fp != NULL )
	    fclose( fp );
	return;
    }

    do 
    {
	len = fread( buffer, 1, BUFFER_SIZE, fp );
	write( client_sockfd, buffer, len );
    } while ( len == BUFFER_SIZE );
	
    fclose( fp );
}
